;
; General SRAM buffering macros for UART reading and writing
; 
;
; Andreas Tnne
; atoenne@t-online.de
;
; No copyright. This is free software. Thank you for naming the author
;
; Web released version 1.0
;
;***********************************************************************


;*********************   USAGE   ***************************************
;
; .DEF Zero						register with a constant zero value (for speed optimizations)
; .DEF UartTmp					register > r15
; .DEF UartBufferRead			index into the buffer where the next byte can be found > r15
; .DEF UartBufferWrite			index into the buffer where the next byte can be written > r15
; .DEF UartBufferStatus			what do we know about the buffer > r15
;
; .SET UART_BAUDRATE
; .SET CLOCK
; .SET UART_BUFFER_SIZE   		max 256 because the pointers into the buffer are single registers
;
; InitializeUART				General setup of the buffer and the UART speed
;								Modifies: UartTmp, UartBufferRead, UartBufferWrite, UartBufferStatus
; UartClearBuffer				Empty the buffer and reset its status
;								Modifies: UartBufferRead, UartBufferWrite, UartBufferStatus
; UartBufferRead REG			Read the next byte from the buffer into REG
;								Modifies: UartBufferRead, UartBufferStatus, Z
; UartBufferWrite REG			Write the byte in REG into the buffer
;								Modifies: UartBufferWrite, UartBufferStatus, Z

; in case of bad rounding errors it is ok to SET UART_UBRR manually after including this file
; however note that this will certainly mean your UART is running off the real speed by more than
; 1% and this means PROBLEMS

.SET	UART_UBRR = (CLOCK / (16 * UART_BAUDRATE)) - 1

;###################################################################

; ****** InitializeUART *********
; initialize the UART speed and setup the buffer

.MACRO InitializeUART
	ldi UartTmp, UART_UBRR
	out UBRR, UartTmp
	clr UartBufferRead
	clr UartBufferWrite
	ldi UartBufferStatus, UART_BUFFER_EMPTY
.ENDMACRO

; *****************************************************************
;
; the buffer is a ring buffer of UART_BUFFER_SIZE
; the current read position is UART_BUFFER + UART_BUFFER_READ
; the current write position is UART_BUFFER + UART_BUFFER_WRITE
; the buffer is EMPTY when both read and write index are equal before or after reading
; the buffer is OVERRUN when both read and write index are equal after writing
; 	in this case we also set the EMPTY flag  to avoid reading fractional data from the buffer

; status bits about the buffer

.EQU	UART_BUFFER_OVERRUN = 1		; set when the buffer cannot take any more 
.EQU	UART_BUFFER_EMPTY = 2		; set when the buffer is empty

.EQU	UART_BUFFER_OVERRUN_BIT = 0
.EQU	UART_BUFFER_EMPTY_BIT = 1

.DSEG
UART_BUFFER: .BYTE UART_BUFFER_SIZE
.CSEG

;******** UartClearBuffer ********
; clear the buffer contents
; note: InitializeUART already includes this
; it is recommended to clear the buffer in a controlled fashion
; when the buffer is OVERRUN

.MACRO UartClearBuffer
	clr UartBufferRead
	clr UartBufferWrite
	ldi UartBufferStatus, UART_BUFFER_EMPTY
.ENDMACRO

;******** UartBufferRead *********
; read the next byte into the register designated by the macro argument
; and increment the READ position

.MACRO _UartBufferRead
	ldi ZL, LOW(UART_BUFFER)
	ldi ZH, HIGH(UART_BUFFER)
	add ZL, UartBufferRead
	adc ZH, Zero								; Z is the SRAM address in the buffer at the read index
	ld @0, Z
	inc UartBufferRead
	cpi UartBufferRead, UART_BUFFER_SIZE
	brne UIR1
	clr UartBufferRead
UIR1:
	cp UartBufferRead, UartBufferWrite
	brne UIR2
	sbr UartBufferStatus, UART_BUFFER_EMPTY
UIR2:
.ENDMACRO

;******** UartBufferWrite *********
; write the byte in the register designated by the first macro argument
; and increment the WRITE position

.MACRO _UartBufferWrite
	ldi ZL, LOW(UART_BUFFER)
	ldi ZH, HIGH(UART_BUFFER)
	add ZL, UartBufferWrite
	adc ZH, Zero								; Z is the SRAM address in the buffer at the write index
	st Z, @0
	inc UartBufferWrite
	cpi UartBufferWrite, UART_BUFFER_SIZE
	brne UIW1
	clr UartBufferWrite
UIW1:
	cbr UartBufferStatus, UART_BUFFER_EMPTY
	cp UartBufferRead, UartBufferWrite
	brne UIW2
	sbr UartBufferStatus, UART_BUFFER_OVERRUN
UIW2:
.ENDMACRO

