;
; General routines for the HD44780 (and compatible) display controller
;
; Andreas Tnne
; atoenne@t-online.de
;
; No copyright. This is free software. Thank you for naming the author
;
; Web released version 1.0
;
;***********************************************************************


;*********************   USAGE   ***************************************
;
; .include "Delay.asm"
;
; .DEF Scratch				General scratch register
; .DEF LCDValue				Value for reading and writing the LCD >= r16 
;
;
; rcall LCDInit				Initial setup of the AVR and synchronize with the LCD and allow sending commands
;							Modifies: LCDValue, DelayInnerLoop, DelayOuterLoop
; rcall LCDWriteCommand		Send LCDValue as a command to the LCD
;							In: LCDValue Modifies: LCDValue
; rcall LCDWriteData		Send LCDValue as a data byte to the LCD
;							In: LCDValue Modifies: LCDValue
; rcall LCDReadAddress		Read the current write address from the LCD
;							Out: LCDValue Modifies: 
; rcall LCDReadData			Read the byte at the current writing address from the LCD
;							Out: LCDValue Modifies: 
; LCDFunctionSet FLAGS		Set number of display lines and font size (call only as first command)
;							Flag LCDMultiLine for more than one display line
;							Flag LCDLargeFont for 5x11 pixel font (5x8 is default) see LCD datasheet is available
;							Modifies: LCDValue
; LCDDisplayOn FLAGS		Turn the display on and set the cursor mode
;							Flag LCDCursorOn switches the cursor on as a steady underline
;							Flag LCDBlockCursor turns the cursor into a blinking block inverting the current character
;							Modifies: LCDValue
; LCDDisplayOff				Turns the display off and hide the cursor
;							Modifies: LCDValue
; LCDClearDisplay			Clears the display and homes the cursor. SLOW!
;							Modifies: LCDValue
; LCDCursorHome				Sets the cursor to the topleft position; undos shifted display. SLOW!
;							Modifies: LCDValue
; LCDSetEntryMode FLAGS		Specify how to change the cursor when a data byte is written
;							Flag LCDIncrement to increment the cursor or decrement otherwise
;							Flag LCDScroll to scroll the display at the edge of the display instead
;								of wrapping to another line (not the logically next line!)
;							Modifies: LCDValue
; LCDMoveCursorRight		Move the cursor and wrap at line breaks
; LCDMoveCursorLeft
; LCDShiftDisplayRight		Shift the display horizontally
; LCDShiftDisplayLeft		
;							The above cursor/display commands modify LCDValue
; LCDSetDisplayAddress		Set the cursor and next write address to LCDValue (7-bit)
;							Modifies: LCDValue
; LCDSetCGAddress ADR		Set the next write address to LCDValue in the character RAM (6-bit)
;							Modifies: LCDValue
;

; Here is a typical sequence of statements that use this driver to initialize a 4x20 LCD display
;
;	rcall LCDInit				; IO-Ports setup and delay for LCD internal boot phase; set 4-bit interface
;	LCDFunctionSet LCDMultiLine	; set LCD to 4-line, 5x8 dots font
;	LCDDisplayOn LCDCursorOn	; display on, steady underscore cursor
;	LCDClearDisplay
;	LCDSetEntryMode LCDIncrement ; increment cursor on write, no scrolling


;********************* ASSUMPTIONS *************************************
;
; Some assumptions about the connection of the LCD to the AVR
; 4-bit access
; the databits are connected to B0-B4
; E (enable) is connected to B5
; R/W is connected to B6
; RS (register select) is connected to B7
;

.SET LCDWritePort=PORTB
.SET LCDReadPort=PINB
.SET LCDDirectionPort=DDRB

; the assumption that Bit0-Bit4 is the data nibble is not configurable
.SET LCDBitE=5			; the E bit triggers data transfer to/from the LCD on the falling edge
.SET LCDBitRW=6			; R/~W = 1 for reading and = 0 for writing to the LCD
.SET LCDBitRS=7			; RS = 1 for reading/writing display data and = 0 for commands

;********************* Timing notes ************************************
;
; The general timing of the HD4470 (and compatible) depends to a large part on 
; the speed of the connected display and the controller system clock. These factors are
; out of the control of the user and hence we must be conservative in the initial
; setup of the controller. We do use polling of the busy flag to avoid needless 
; delays that wait for the controller to finish a write. 
; See LCDEnableWrite and LCDEnableRead for additional timing assumptions:
; Figures:
; The typical controller clock is 250kHz.
; The minimal enable cycle length is 1us. E must be high for at least 0.45us. This is roughly
;	equivalent to two AVR clock cycles at 4Mhz. (Hence the three nop in the LCDEnableX functions)
; The address setup time (necessary time between setting RS, R/W and running the enable cycle) is 
; neglectable but note that we must not set all three at the same time. First set RS and R/W with a
; separate instruction and then run the enable cycle.

;#######################################################################

; Port DDR masks

.SET LCDWriteOperationMask=(1<<LCDBitRS) | (1<<LCDBitRW) | (1<<LCDBitE) | 0b1111
.SET LCDReadOperationMask=(1<<LCDBitRS) | (1<<LCDBitRW) | (1<<LCDBitE)

;******* LCDAllowWrite ***********
; switch the interface to allow writing to the LCD (default mode)

.MACRO LCDAllowWrite
	ldi Scratch, LCDWriteOperationMask
	out LCDDirectionPort, Scratch
.ENDMACRO

;******* LCDAllowRead ***********
; switch the interface to allow reading from the LCD
; NOTE: it is expected that after reading from the LCD the interface
; is reset to writing since writing is a much more common operation

.MACRO LCDAllowRead
	ldi LCDValue, LCDReadOperationMask
	out LCDDirectionPort, LCDValue
.ENDMACRO

;******* LCDInit *******
; Perform the initial setup before we may send commands to the LCD
; configure the output port and wait for 30ms until the LCD is ready
; The datasheets require only 15ms but since this only happens at the 
; power on time it does not hurt to be on the safe side.
; Note: this routine must be called early in the reset routine and strictly
; 		before any data is sent to the LCD. Orelse you might spook your LCD :-)

LCDInit:
	LCDAllowWrite

	Delay 30000


; Run the LCD initialization phase 
; set the LCD to 4bit data bus
; the required further configuration like font, cursor, display on can be set using 
; the regular command macros after LCDInit is finished.
; The initialization phase consist of three dummy function set commands (only the higher nibble
; is relevant) for synchronization followed by another function set command (again only the
; higher nibble is read) which determines the data length DL to be 4-bit. Once the DL is set
; we have to send each byte as higher and lower nibble separate on the AVR port. 
; We are very conservative with delay timing here to assure that the initialization goes through
; independent of the actual LCD timing.

	ldi LCDValue, 0b0011		; function set command (it doesnt matter that it specifies DL=8bit) 
	rcall LCDEnableWrite
	Delay 5000					; datasheet requires > 4.1ms, we wait 5ms
	
	rcall LCDEnableWrite
	Delay 5000		

	rcall LCDEnableWrite
	Delay 5000

	ldi LCDValue, 0b0010		; function set command, DL=0 means 4-bit interface
	rcall LCDEnableWrite
	Delay 5000

	; from this point on we read the BUSY status of the LCD instead of the delays
	; the next command MUST BE a LCDFunctionSet to determine number of display lines and font
	ret


;******* LCDEnableWrite **********
; Write the lower nibble of LCDValue to the output port
; Note: the bits of the upper nibble determine the RS and R/W status
; Trigger the E line once to write the status of the output port to the LCD
; Timing note: the E line must be high for at least 450ns. This time is 
; guaranteed by the three nop instructions at 4MHz.
; At 10Mhz you need to insert 11 nop
; At 1Mhz you need no nop

LCDEnableWrite:
	out LCDWritePort, LCDValue
	sbi LCDWritePort, LCDBitE
	nop
	nop
	nop
	cbi LCDWritePort, LCDBitE
	ret

;******* LCDEnableRead **********
; Trigger the E line once and read one nibble into LCDValue while the
; line is high. The same timing considerations as for LCDEnableWrite apply.

LCDEnableRead:
	sbi LCDWritePort, LCDBitE
	nop
	nop
	nop
	in LCDValue, LCDReadPort
	cbi LCDWritePort, LCDBitE
	andi LCDValue, 0x0F
	ret

;******** LCDWriteCommand ***********
; Write the contents of LCDValue as a command to the LCD
; The byte is sent in the order upper / lower nibble
; After the second nibble we poll the busy flag to return as early as possible

LCDWriteCommand:
	push LCDValue
	swap LCDValue
	andi LCDValue, 0x0F			; note: RS=0, R/W=0 => write command
	rcall LCDEnableWrite		; send the upper nibble
	pop LCDValue
	andi LCDValue, 0x0F

LCDWriteAndWait:
	rcall LCDEnableWrite		; send the lower nibble
LCDCommandBusy:
	rcall LCDReadBusyFlag
	brcs LCDCommandBusy			; test if BUSY set
	ret

;******* LCDWriteData ***********
; Write the contents of LCDValue as a data byte to the LCD
; The byte is sent in the order upper / lower nibble
; After the second nibble we poll the busy flag to return as early as possible
; Note: this increments the address and cursor as set by the LCDSetEntryMode command

LCDWriteData:
	push LCDValue
	swap LCDValue
	andi LCDValue, 0x0F
	ori LCDValue, 1<<LCDBitRS	; note: RS=1, R/W=0 => write data
	rcall LCDEnableWrite		; send the upper nibble
	pop LCDValue
	andi LCDValue, 0x0F
	ori LCDValue, 1<<LCDBitRS
	rjmp LCDWriteAndWait


;****** LCDReadBusyFlag *******
; Read the busy flag into the carry flag C
; Test with brcs LCDValue afterwards
; Timing note: we do not need to optimize this busy flag polling
; The fastest write operation to the LCD takes 37us typically while this
; routine takes aprox 8us at 4MHz. So we may expect five polling iterations
; before BUSY goes down and an average slack of 3.5us


LCDReadBusyFlag:
	rcall LCDReadCommand
	rol LCDValue				; move bit 7 (BUSY flag) into carry
	ret

;******* LCDReadAddress ********
; Read the current display address position (used for next LCDWriteData) into LCDValue

LCDReadAddress:
	rcall LCDReadCommand
	andi LCDValue, 0b01111111	; mask out BUSY flag
	ret

; private routine for reading the busy flag and adress
; the dual routine to LCDreadData

LCDReadCommand:
	cbi LCDWritePort, LCDBitRS	; R/W=1 means read, RS=0 means command

LCDRead:
	LCDAllowRead
	sbi LCDWritePort, LCDBitRW	

	rcall LCDEnableRead
	swap LCDValue				; store as upper nibble
	mov Scratch, LCDValue
	rcall LCDEnableRead
	or LCDValue, Scratch		; join both nibble
	LCDAllowWrite
	ret

;******* LCDReadData *******
; Read the byte at the current cursor point into LCDValue
; Note: this increments the address as set by the LCDSetEntryMode command
; It does not increment the cursor position!

LCDReadData:
	sbi LCDWritePort, LCDBitRS	; R/W=1 means read, RS=1 means data
	rcall LCDRead
	push LCDValue
	rcall LCDCommandBusy
	pop LCDValue
	ret

;****** LCDFunctionSet *******

.SET LCDMultiLine=0b1000		; more than one display line
.SET LCDLargeFont=0b0100		; 5x11 font (very unusual)

; COMMAND: Function set
; @0 = command mask nf??
;		n = 1 Display more than 1 line or else n = 0
;		f = 1 5x11 dots font, f = 0 5x8 dots font (normal)

.MACRO LCDFunctionSet
	ldi LCDValue, 0b00100000		; assumption: 4-bit data interface
	ori LCDValue, @0
	rcall LCDWriteCommand
.ENDMACRO

;****** LCDDisplayOn ******

.SET LCDCursorOn=0b10				; a cursor is shown
.SET LCDBlockCursor=0b01			; the cursor is blinking the current character; otherwise its a steady underline

; COMMAND: Display on
; @0 = cursor activation mask cb
;		c = 1 for cursor on
;		b = 1 for blinking cursor on

.MACRO LCDDisplayOn
	ldi LCDValue, 0b00001100
	ori LCDValue, @0
	rcall LCDWriteCommand
.ENDMACRO

;****** LCDDisplayOff *******
; COMMAND: Display off
;		also disables cursor and blinking

.MACRO LCDDisplayOff
	ldi LCDValue, 0b00001000
	rcall LCDWriteCommand
.ENDMACRO

;****** LCDClearDisplay *******
; COMMAND: Clear display
; Note: this is a pretty slow operation (typically 1.52ms)

.MACRO LCDClearDisplay
	ldi LCDValue, 1
	rcall LCDWriteCommand
.ENDMACRO

;****** LCDCursorHome *******
; COMMAND: Cursor home
; Note: this is a pretty slow operation (typically 1.52ms)

.MACRO LCDCursorHome
	ldi LCDValue, 2
	rcall LCDWriteCommand
.ENDMACRO

;****** LCDSetEntryMode ******

.SET LCDIncrement=0b10
.SET LCDScroll=0b01

; COMMAND: Set entry mode
; @0 = is
;		i = 1 for incrementing the write address
;		s = 1 for scrolling display in the i direction instead of moving the cursor

.MACRO LCDSetEntryMode
	ldi LCDValue, 4
	ori LCDValue, @0
	rcall LCDWriteCommand
.ENDMACRO

;***** LCDMoveCursorRight *****
; COMMAND: Move cursor right (no scroll)

.MACRO LCDMoveCursorRight
	ldi LCDValue, 0b00010100
	rcall LCDWriteCommand
.ENDMACRO

;***** LCDMoveCursorLeft *****
; COMMAND: Move cursor left (no scroll)

.MACRO LCDMoveCursorLeft
	ldi LCDValue, 0b00010000
	rcall LCDWriteCommand
.ENDMACRO

;***** LCDShiftDisplayRight *****
; COMMAND: Shift the display right one character position

.MACRO LCDShiftDisplayRight
	ldi LCDValue, 0b00011100
	rcall LCDWriteCommand
.ENDMACRO

;***** LCDShiftDisplayLeft *****
; COMMAND: Shift the display left one character position

.MACRO LCDShiftDisplayLeft
	ldi LCDValue, 0b00011000
	rcall LCDWriteCommand
.ENDMACRO

;***** LCDSetDisplayAddress ******
; COMMAND: Display RAM Adress to LCDValue
; LCDValue = 7-bit address
; Note: setting this address also determines that all subsequent LCDWriteData bytes are sent to 
; the display RAM

.MACRO LCDSetDisplayAddress
	ori LCDValue, 0b10000000
	rcall LCDWriteCommand
.ENDMACRO

;***** LCDSetCGAddress ******
; COMMAND: Character generator RAM Adress to LCDValue
; LCDValue = 6-bit address
; Note: setting this address also determines that all subsequent LCDWriteData bytes are sent to
; the character RAM

.MACRO LCDSetCGAddress
	ori LCDValue, 0b01000000
	rcall LCDWriteCommand
.ENDMACRO

