; 
; MIDI Monitor Event Buffer
; 
; Routines for the buffer of received and filtered MIDI events
;
;
; Andreas Tnne
; atoenne@t-online.de
;
; No copyright. This is free software. Thank you for naming the author
;
; Web released version 1.0

; ******************** Physical Buffer Organization *****************************************
;
; The buffer is organized as a ring buffer when the monitor is in continuous mode
; and as a normal limited buffer when the monitor is in sample&hold mode
;
; The buffer has a start BUFFER_START and a size BUFFER_SIZE
; The window that is displayed by the LCD is the bytes range DisplayTop to DisplayTop + DisplaySize - 1
; The current writing position is BufferWrite
;
; The buffer is cleared with the contents of register Empty (typically the active sensing event 0xFE)

; NOTE: the buffer size must be a multiple of six. Otherwise BufferWrite will not operate correctly.

; ******************** Logical Buffer Organization ******************************************
;
; The buffer contains MIDI events. Events are 1-3 or variable bytes in size and we only record and show
; completely received events. Three features of the MIDI protocol complicates the buffer management:
; 1. Realtime events (always one byte) can occur anytime even interleaved with other multi byte events
; 2. System exclusive event are of variable size and typically much larger than the SRAM of the AVR controllers
; 3. Consecutive similar regular events (same type, same channel) can be abreviated by omitting the 
;	event status byte. This is called running status.
; The majority of events will be regular events like note on, note off, controller values.
; To deal with these special cases we make a few simplifying choices. All off these can be lifted but that
; will need a much more complicated programming.
;
; - SystemExclusive events are dropped by the UART receiver. The only interesting parts that we would be able to show
;	are the SysEx header, which indicates the device manufacturer and possible some status information. 
; - ActiveSensing is filtered out and we use the event byte 0xFE as a buffer filler instead (normally).
; - All runnning status is automatically expanded to the full regular event shape by the receiver routine. This *is*
;	a deliberate modification of the received data but one that should be transparent anyway. We cannot imagine a
;	situation where this expansion of running status would be problematic. In particular since the MIDI-Thru is done
;	at the circuit level and not touched by our programming.
; - Since we expand running status and since most events will be thus three byte events, we reserve three bytes for
;	each received event in the buffer regardless of the actual event size. Events with fewer bytes will be padded by FE.
; - Realtime events that are interleaved with another event are buffered immediately BEFORE the other event.
;
; The following rules describe the operation of the buffer (expanding and event filtering is done elsewhere)
; We distinguish the modes HEX and SYMBOLIC for displaying and CONTINUOUS vs. SAMPLE&HOLD for buffer wrap.
;
; CONTINUOUS mode means that writing the buffer wraps around the buffer limit and the newer bytes overwrite the oldest
; bytes. Also the LCD window may wrap around the buffer limit and may collide with the write position. The implications
; of this are detailed below.
; SAMPLE&HOLD mode means that the buffer is filled from left to right and then reading stops (actually all events are
; filtered out). The writing and LCD window does not wrap.
; 
; HEX display mode shows the buffer contents uninterpreted (but with running status expanded) in four rows with
; six bytes each. The padding Empty event bytes are *not* displayed but shown as spaces.
; SYMBOLIC display mode shows the buffer contents with their symbolic event names and parameters each on its own line.
; The LCD display window is always of this size (24 bytes or 12 bytes for 8 or 4 events) even if the buffer is not full.
; Scrolling occurs either by steps of six or three bytes.
;
; CONTINOUS buffer filling allows for the write position to overrun the LCD window top (which means the LCD window shows some
; bytes that are now expired and overwritten by the most recent received bytes). In this case the LCD window is
; scrolled down and the bytes of the top line (6 or 3) are filled with 0xFE before the next byte can be written to the
; buffer. Also if the write position was in the last line of the LCD window the writing to the next line will drag the window
; along to show the most current events. 
;
; Scrolling the LCD window is limited to the top and to the bottom by the current write position. Observe the 
; necessary special cases when the write position is within the LCD window or when the write position has not
; yet wrapped around at least once (then we would scroll up into the empty buffer). This special case is indicated by
; the flag BufferUsed. It is reset when the buffer is cleared and it is constantly 0 in SAMPLE&HOLD. When BufferUsed is 0
; then the buffer top is 0 and it is BufferWrite when BufferUsed is 1.
;
; The DisplayTop and DisplaySize values are always multiple of three resp. six. This has to be respected when switching 
; the display mode from SYMBOLIC to HEX.


; ******************** USAGE *************************
;
; .DEF DisplayMode						0 for HEX display, 1 for SYMBOLIC display
; .DEF BufferMode						0 for CONTINUOUS, 1 for SAMPLE&HOLD
; .SET BUFFER_SIZE						The desired size of the MIDI buffer
; .DEF BufferWrite						The position in the buffer that will be written next
; .DEF DisplayTop						The first displayed byte
; .DEF DisplaySize						The number of displayed bytes
; .DEF LineSize							The number of bytes in one line
; .DEF BufferTmp						For various tasks; unmodified by the buffer routines
; .DEF MIDIValue						Value to buffer; from the receiver
; .DEF BufferUsed						1 if the buffer is filled completely and the BufferWrite has wrapped around at least once.
;										This flag is 0 after a clear buffer and in SAMPLE&HOLD mode.
; register Y is reserved for the buffer access
;
; rcall BufferClear						Clear the buffer
;										Modifies: 
; BufferIncIndex REG					Increments the register value within the buffer bounds
; BufferDecIndex REG					Decrements the register value within the buffer bounds
; rcall BufferWrite REG					Write the byte in REG into the buffer and scrolls/wraps as necessary


; ####################################################


.DSEG
BUFFER_START:
.byte BUFFER_SIZE
BUFFER_END:
.CSEG

;***** BufferClear *****
; Clears the buffer with the empty event

BufferClear:
	ldi Scratch, BUFFER_SIZE
	push BufferTmp
	ldi YL, LOW(BUFFER_START)
	ldi YH, HIGH(BUFFER_START)
	mov BufferTmp, Empty
BufferClearLoop:
	st Y+, BufferTmp
	dec Scratch
	brne BufferClearLoop
	pop BufferTmp
	clr BufferWrite					; set the write position to the top of the buffer
	clr BufferUsed
	ret

;***** BufferToggleMode *****
; Toggle between CONTINUOUS to SAMPLE&HOLD 

BufferToggleMode:
	ldi Scratch, 1
	eor BufferMode, Scratch			; toggle the bit
	mov Scratch, BufferMode
	com Scratch
	lsl Scratch
	out PORTD, Scratch				; show the new LED status
	rcall BufferClear
	rjmp DisplayClear

;**** ScrollWindowDown ******
; Scroll the current displayed window down one line
; Confine the window to the buffer size. i.e. display end may not cross the current write position

ScrollWindowDown:
	add DisplayTop, LineSize				; move the display window by one line
	rcall DisplayPos						; determine the new index of the window into the buffer
	add Scratch, DisplaySize				; = first byte index after the window	
	cpi Scratch, BUFFER_SIZE				; check if we are outside the buffer and crossed the write position
	brmi ScrollLineDownOK					; index after window < buffer size?
	subi Scratch, BUFFER_SIZE
	sub DisplayTop, Scratch					; if yes, reduce the scrolling to the maximum
ScrollLineDownOK:
	cpi DisplayTop, BUFFER_SIZE				; now wrap the window top if necessary
	brmi ScrollWindowDown1
	subi DisplayTop, BUFFER_SIZE
ScrollWindowDown1:
	ret

;***** ScrollWindowUp ********
; Scroll the current displayed window up one line
; Confire the window to the buffer size. i.e. display top may not cross the current write position

ScrollWindowUp:
	rcall DisplayPos						; determine the index of the window into the buffer
	cp Scratch, LineSize					; less than one line size space?
	brmi ScrollLineUpShort					; if yes then only move the delta instead of a full line
	sub DisplayTop, LineSize				; move the display window by one line
	rjmp ScrollLineUpOK
ScrollLineUpShort:
	sub DisplayTop, Scratch					; move the display window to the topmost position
ScrollLineUpOK:
	cp DisplayTop, Zero						; did we shoot past the top? Then wrap around
	brpl ScrollWindowUp1					; no wrap needed
	subi DisplayTop, -BUFFER_SIZE
ScrollWindowUp1:
	ret
;***************************************************************************************

;********* IncBufferIndex REG ***************
; Increment the buffer index in REG with wrap around the buffer size

.MACRO BufferIncIndex
	cpi @0, BUFFER_SIZE-1
	brne BufferIncIndex1
	ldi @0, -1
BufferIncIndex1:
	inc @0
.ENDMACRO

;********* DecBufferIndex REG **************
; Decrement the buffer index in REG with wrap around the buffer size

.MACRO BufferDecIndex
	cpi @0, 0
	brne BufferDecIndex1
	ldi @0, BUFFER_SIZE
BufferDecIndex1:
	dec @0
.ENDMACRO

;******** BufferLoadPointer REG *****************
; Load BufferPointer with the pointer in SRAM of the current buffer position REG

.MACRO BufferLoadPointer
	ldi BufferPointerL, LOW(BUFFER_START)
	ldi BufferPointerH, HIGH(BUFFER_START)
	add BufferPointerL, @0
	adc BufferPointerH, Zero
.ENDMACRO

;******* BufferTop **********
; Load Scatch with the top index of the current buffer
; Note: this is the oldest byte in the buffer and will be overwritten with the next byte

BufferTop:
	cp BufferUsed, Zero
	brne BufferTopIsWrite
	clr Scratch
	ret
BufferTopIsWrite:
	mov Scratch, BufferWrite
	ret


;******* DisplayPos **********
; Load Scratch with the number of bytes of the DisplayTop into the current buffer

DisplayPos:
	rcall BufferTop						; Scratch is the index of the buffer top position (oldest byte)
	mov BufferTmp, Scratch
	mov Scratch, DisplayTop
	sub Scratch, BufferTmp
	brmi DisplayPosNegative
	ret
DisplayPosNegative:						; the display top lies before the buffer top; a result of wrapping
	subi Scratch, -BUFFER_SIZE			; correct the wrapping
	ret

;****************************************************************************************

;********** BufferWrite  ***************
; Write the byte in register MIDIValue into the buffer and scroll down the display window if the write
; position and display window collided or if the write position left the display window (we drag the display
; window after the write position in this case)
; Note: we do not trigger displaying here. This is the task of the receiver routine which knows
; when a complete event was received.

_BufferWrite:
	cpi BufferWrite, BUFFER_SIZE-1			; we need to wrap around?
	brne BufferWriteNoWrap
	inc BufferUsed							; then remember that the whole buffer is in use now
BufferWriteNoWrap:
	cp BufferUsed, Zero
	breq BufferWriteNoTopTest				; do not test collision with DisplayTop unless we wrapped around

	sbrs BufferMode, 0						; SAMPLE&HOLD mode?
	rjmp BufferWriteTopTest					; no -> continue usual operation
	cp BufferWrite, Zero					; yes -> check if the buffer is filled completely
	brne BufferWriteTopTest					; buffer still not filled
	ret										; buffer is filled -> stop recording

BufferWriteTopTest:
	cp BufferWrite, DisplayTop				; writing new bytes into the old displayed values? 
	breq BufferWriteScroll

BufferWriteNoTopTest:
	mov Scratch, DisplayTop
	add Scratch, DisplaySize				; = first byte position after the current display window							
	cpi Scratch, BUFFER_SIZE
	brmi BufferWrite1
	subi Scratch, BUFFER_SIZE
BufferWrite1:
	cp BufferWrite, Scratch					; or moving the write position outside the LCD window?
	brne BufferWriteByte

BufferWriteScroll:
	rcall BufferScrollWindowDown			; then move the display window one line down

BufferWriteByte:
	BufferLoadPointer BufferWrite			; Y = pointer to next to be written position
	st Y, MIDIValue
	BufferIncIndex BufferWrite				; BufferWrite = index of next to be written position
	ret

;***************************************************************************************

;********* BufferScrollWindowDown *************
; Scroll the display window one logical line down and clear the new line

BufferScrollWindowDown:
	add DisplayTop, LineSize				; move the display window by one line
	cpi DisplayTop, BUFFER_SIZE
	brne BufferScrollWindowDown1
	clr DisplayTop
BufferScrollWindowDown1:
	BufferLoadPointer BufferWrite
	mov Scratch, LineSize
BufferScrollWindowLoop:						; now erase the new space
	st Y+, Empty
	dec Scratch
	brne BufferScrollWindowLoop
	ret

;*****************************************************************************************
