; 
; MIDI Monitor
; 
; Records a limited number of MIDI events received on its serial interface and
; displays these events in various formats on a 4x20 LCD
;
;
; Andreas Tnne
; atoenne@t-online.de
;
; No copyright. This is free software. Thank you for naming the author
;

; Web released version 1.0

.include "2313def.inc"
.listmac

;***************************************************************
; general setup

; note: we can reuse some registers for multiple purposes (since the register >=16 are rare) but
; we must not double use the registers for the UART buffer as these are modified in the receive
; interrupt handler. 

.DEF Empty=r1									; the empty data byte 0xFE; can be any other realtime event byte if active sensing should be received
.DEF _Byte1=r2									; first byte of the received event
.DEF _Byte2=r3									; second byte  of the received event or Empty for 1-byte events
.DEF _Byte3=r4									; third byte of the received event or Empty for 2-byte events
.DEF LastStatus=r5								; the last status byte or 0 if unknown
.DEF BytesToGo=r6								; the number of remaining bytes for this event or negative for Sysex (variable size)
.DEF DisplayScratch=r7							; only for the event displaying routines
.DEF BufferUsed=r9								; 1 if the buffer is filled completely and 0 if writing has not yet wrapped around
.DEF LineSize=r10								; number of bytes in one display line
.DEF EventType=r11								; communication value between event receiver and filter
.DEF Zero=r15									; must NOT be r0 because Zero is used in interrupt handlers and r0 is modified by the LPM instruction
.DEF Tmp=r16									; all Tmp registers must be preserved by subroutines
.DEF UartTmp=r16
.DEF LCDCursorTmp=r16							; must be distinct from LCDValue
.DEF BufferTmp=r16
.DEF DisplayTmp=r16
.DEF Scratch=r17								; may be modified by subroutines adlib
.DEF LCDValue=r18								; parameter to the LCD routines
.DEF MIDIValue=r19								; current MIDI bytes in the processing
.DEF BUTTONValue=r19							; for the button reading
.DEF EEPromValue=r19							; IO buffer for EEPROM reading or writing
.DEF UartBufferRead = r20						; index in the UART receive buffer for reading
.DEF UartBufferWrite = r21						; index in the UART receive buffer for writing
.DEF UartBufferStatus = r22						; status of the UART receive buffer (empty? overflow?)
.DEF DisplayTop = r23							; first byte in buffer that is displayed
.DEF DisplaySize = r24							; number of bytes in the buffer that are displayed
.DEF BufferWrite = r25							; the byte that will be written next in the buffer
.DEF BufferMode = r26							; CONTINOUS or SAMPLE&HOLD buffer?
.DEF DisplayMode = r27							; HEX or SYMBOLIC display? Display deactivated?
.DEF EEPromAddressL = r28						; Pointer into EEProm
.DEF BufferPointerL = r28
.DEF BufferPointerH = r29
.DEF DelayInnerLoop=r30							; delay timing parameters - preserve Z when delays are used
.DEF DelayOuterLoop=r31							; 

.SET CLOCK = 4000000							; Hz
	
.SET UART_BAUDRATE = 31250
.SET UART_BUFFER_SIZE = 16

.SET BUFFER_SIZE = 0x5A							; = 90 ; must be a multiple of 6

.SET BUTTONPort = PIND
.SET BUTTONMask = 0b01111100					; bit7 is not available on the 2313, bit0-1 are for UART

;***************************************************************

.CSEG

.EQU	RESET=0

.ORG RESET
	rjmp Main				; RESET
	rjmp RESET				; IRQ0 handler
	rjmp RESET				; IRQ1 handler
	rjmp RESET				; Timer1 capture handler
	rjmp RESET				; Timer1 Compare handler
	rjmp RESET				; Timer1 Overflow handler
	rjmp RESET				; Timer0 Overflow handler
	rjmp UARTReceive		; UART RX complete handler
	rjmp RESET				; UDR empty handler
	rjmp RESET				; UART TX complete handler
	rjmp RESET				; Analog comparator handler


;***************************************************************

.include "Basic Tools.asm"
.include "Delay.asm"
.include "LCD Driver.asm"
.include "LCD Cursor 4x20.asm"
.include "LCD Text.asm"
.include "Button.asm"
.include "EEProm.asm"
.include "BufferedUART.asm"
.include "BufferedUART Handler.asm"
.include "MIDI Monitor Messages.asm"
.include "MIDI Monitor Buffer.asm"
.include "MIDI Monitor Receiver.asm"
.include "MIDI Monitor Display.asm"

;***************************************************************

Main:
	SetSmallStack RAMEND
	ldi Tmp, ACTIVESENSING
	mov Empty, Tmp				; define the empty buffer contents => this is event ACTIVESENSING that is not used elsewhere
	clr Zero
	clr DisplayMode				; HEX display
	clr BufferMode				; CONTINUOUS buffer recording
	clr LastStatus				; we have not seen any MIDI event yet
	clr BytesToGo
	clr BufferUsed				; buffer is not filled; buffer top is 0 instead of BufferWrite

; now do some basic sanity check for the EEPROM contents
; since address 0 tends to get corrupted by brown outs mot frequently we write a magic byte
; and check if it still present on the next reset

	ldi EEPromAddressL, LOW(MAGIC)	; read the magic byte to see if the EEPROM needs initialization
	rcall EEPromRead
	cpi EEPromValue, MAGICBYTE
	breq InitLCD
	rcall EEPromInit

InitLCD:
	rcall LCDInit				; IO-Ports setup and delay for LCD internal boot phase; set 4-bit interface
	LCDFunctionSet LCDMultiLine	; set LCD to 4-line, 5x8 dots font
	LCDDisplayOn LCDCursorOn	; display on, steady underscore cursor
	LCDClearDisplay
	LCDSetEntryMode LCDIncrement ; increment cursor on write, no scrolling

	rcall Greetings

	InitializeUART				; set the speed of the UART - reading/writing is determined later
	InitializeUARTReading		; now we handle interrupts coming from the UART
	rcall BufferClear			; clear the buffer and reset the write position
	rcall DisplayClear			; reset the display window and contents
	sei							; start receiving MIDI bytes

	ldi Scratch, 0b00000010
	out DDRD, Scratch			; configure port D for input except for pin 1 (for buffer mode display)
	out PORTD, Scratch			; and set the LED OFF! (meaning CONTINUOUS mode)

Readloop:
	sbrs UartBufferStatus, UART_BUFFER_OVERRUN_BIT			; test for buffer overrun
	rjmp BufferNotOverrun
	rcall Overrun
	rjmp Readloop
BufferNotOverrun:
	sbrc UartBufferStatus, UART_BUFFER_EMPTY_BIT			; test for data in the UART buffer
	rjmp BufferEmpty
	_UartBufferRead MIDIValue								; get this data and
	rcall ReceiveMIDIValue									; process it in the receiver
BufferEmpty:												; UART buffer is empty -> check if function keys are pressed
	rcall BUTTONReadWait									; read the button state and wait until the buttons are released
	sbrc BUTTONValue, 2
	rjmp ToggleDisplayMode
	sbrc BUTTONValue, 3
	rjmp ScrollUp
	sbrc BUTTONValue, 4
	rjmp ScrollDown
	sbrc BUTTONValue, 5
	rjmp Configuration
	sbrc BUTTONValue, 6
	rjmp ToggleBufferMode
	rjmp Readloop


ToggleDisplayMode:
	rcall DisplayToggleMode

Redisplay:
	rcall DisplayRedisplay
	rjmp Readloop

ToggleBufferMode:
	rcall BufferToggleMode
	rjmp Readloop

ScrollUp:
	rcall ScrollWindowUp
	rjmp Redisplay

ScrollDown:
	rcall ScrollWindowDown
	rjmp Redisplay

Configuration:
	rcall Configure
	rjmp Redisplay
